/***************************************************************/
/*                         eml.c                               */
/*　분류지식을 기계학습하는 프로그램  　                       */
/*　　입력은 표준입력으로 주고 출력은 표준출력으로 출력합니다  */
/* 사용방법                                                    */
/*       $./eml 　< (입력파일명）> (출력파일명)                */
/*　입력파일에서는 속성과 카테고리를 0/1로 기술합니다          */
/*　출력파일에는 학습결과를 표시합니다                         */
/***************************************************************/

/* 헤더 파일 포함 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* 기호 상수 정의 */
#define TEXTLENGTH 256 /* 텍스트의 최대길이 */
#define DATANO 256   /* 학습 데이터 세트의 최대 개수 */
#define ATTNO 4 /* 속성값의 개수 */
#define TERMNO 3 /* 논리식의 논리합 항의 개수*/
#define YES 1   /* yes에 대응하는 정수값 */
#define NO 0    /* no에 대응하는 정수값 */
#define SEED 65535    /* 난수의 시드 */
#define REPNO 1000    /* 분류지식의 생성회수*/
#define THVAL 0.7     /* 결과출력의 하한값 */

/* 함수의 프로토타입 선언 */
int readdata(int dataset[DATANO][ATTNO+1]) ;
                     /* 학습 데이터 읽어들이기 */
int rnd() ; /*    0/1 난수 생성   */
void genformula(int formula[TERMNO][ATTNO*2]) ;
                                /* 논리식 생성 */
double evalformula(int formula[TERMNO][ATTNO*2]
  ,int dataset[DATANO][ATTNO+1],int no) ;/* 논리식 평가 */
void purformula(int formula[TERMNO][ATTNO*2]) ; /* 논리식 출력 */
int evals(int formula[TERMNO][ATTNO*2]
      ,int subset[ATTNO+1]); /* 특정 사례에 대한 평가*/
int notval(int v) ;/* 진리값 반전 */

/*******************/
/*    main() 함수  */
/*******************/
int main(int argc,char *argv[])
{
 int no ; /* 학습 데이터 세트의 개수 */
 int i ; /* 반복의 제어문자 */
 int dataset[DATANO][ATTNO+1]={0} ;
       /* 학습 데이터 세트의 속성과 카테고리를 저장 */
 int formula[TERMNO][ATTNO*2]={0} ;
       /* 획득대상이 되는 논리식 */
 double score=0 ;/* 논리식의 평가값 */

 /* 난수의 초기화 */
 srand(SEED) ;

 /* 학습 데이터 읽어들이기 */
 no=readdata(dataset) ;

 /* 논리식의 생성과 검사 */
 for(i=0;i<REPNO;++i){
  genformula(formula) ;/* 논리식 생성 */
  score=evalformula(formula,dataset,no) ;/* 논리식 평가 */
  /* 논리식 출력 */
  if(score>THVAL){
	  printf("\n스코어: %lf\n",score) ;
   purformula(formula) ;
  }
 }
 
 return 0 ;
}

/************************/
/*   readdata()함수     */
/*학습 데이터 읽어들이기*/
/************************/
int readdata(int dataset[DATANO][ATTNO+1])
{
 int no=0 ; /* 학습 데이터 세트의 개수 */
 int att=0 ; /* 학습 데이터 세트의 속성 */

 /* 데이터 입력 */
 while(scanf("%d",&dataset[no][att])!=EOF){
  ++ att ;
  if(att>ATTNO){/* 다음 데이터 */
   att=0 ;
   ++no ;
  }
 }
 return no ;
}


/************************/
/*     rnd()함수        */
/*    0/1 난수 생성     */
/************************/
int rnd()
{
 int rndno ;/* 생성한 난수 */

 while((rndno=((double)rand()/RAND_MAX)*2)==2) ;

 return rndno;
}

/************************/
/*   genformula()함수   */
/*    논리식 생성       */
/************************/
void genformula(int formula[TERMNO][ATTNO*2])
{
 int i,j ;/* 루프 제어 변수 */

 for(i=0;i<TERMNO;++i)
  for(j=0;j<ATTNO*2;++j)
   /* 0 또는1의 난수에 의해 값을 결정 */
   formula[i][j]=rnd() ;
}

/************************/
/*   putformula()함수   */
/*    논리식 출력       */
/************************/
void purformula(int formula[TERMNO][ATTNO*2]) 
{
 int i,j ;/* 루프 제어 변수 */

 for(i=0;i<TERMNO;++i){
  for(j=0;j<ATTNO*2;++j)
   printf("%d ",formula[i][j]) ;
  printf("\n") ;
 }
}
 

/************************/
/*   evalformula()함수  */
/*   논리식 평가        */
/************************/
double evalformula(int formula[TERMNO][ATTNO*2]
  ,int dataset[DATANO][ATTNO+1],int no)
{
 int i ;/* 루프 제어 변수 */
 int point=0 ;/* 정답의 개수 */

 /* 학습 세트의 각 사례에 논리식을 적용 */
 for(i=0;i<no;++i){
  point+=evals(formula,dataset[i]) ; 
 }

 return (double)point/no ;/* 평가값을 반환한다 */
}

/************************/
/*    evals()함수       */
/*특정 사례에 대한 평가 */
/************************/
int evals(int formula[TERMNO][ATTNO*2]
  ,int subset[ATTNO+1])
{
 int orval,andval,result ;/* 계산도중 경과값 */
 int i,j ;/* 루프 제어 변수 */

 andval=YES ; 
 /* 논리곱으로 항을 연결한 식 전체의 평가 */
 for(i=0;i<TERMNO;++i){
  orval=NO ;
  /* 논리합으로 연결한 항의 평가 */
  for(j=0;j<ATTNO*2;++j){
   if((j%2)==0) orval+=subset[j/2]*formula[i][j] ;
   else orval+=notval(subset[j/2])*formula[i][j] ;
  }
  if(orval==NO) andval=NO ;
  else andval*=YES ; 
 }

 if(andval==subset[ATTNO]) return 1 ;/* 결과 일치 */
 else return 0 ;/* 불일치 */
}


/************************/
/*   notval()함수       */
/*   진리값 반전        */
/************************/
int notval(int v)
{
 if(v==YES) return NO ;
 else return YES ;
}

